import json

import numpy as np
from drax_ecc import crypto

from drax_sdk.model.node import State

DRAX_PUBLIC_KEY = "bb099d6dce1a953c7c5d2380815ee02ea191b39206000000ceefb3c222b480459556ce440379cef89db0ccfc04000000"


def encode_state(private_key: bytes, state: State | dict) -> bytes:
    json_data = json.dumps(state.to_dict() if isinstance(state, State) else state)
    json_data_np = np.frombuffer(json_data.encode(), dtype=np.uint8)
    private_key_np = np.frombuffer(private_key, dtype=np.uint8)
    public_key: bytes = bytes.fromhex(DRAX_PUBLIC_KEY)
    public_key_np = np.frombuffer(public_key, dtype=np.uint8)
    encrypted_data_np = crypto.sign(
        private_key_np,
        public_key_np,
        json_data_np,
    )
    encrypted_data_bytes = encrypted_data_np.tobytes()

    return encrypted_data_bytes


def decode_configuration(private_key: bytes, configuration: bytes) -> dict:
    configuration_np = np.frombuffer(configuration, dtype=np.uint8)
    private_key_np = np.frombuffer(private_key, dtype=np.uint8)
    public_key: bytes = bytes.fromhex(DRAX_PUBLIC_KEY)
    public_key_np = np.frombuffer(public_key, dtype=np.uint8)
    decrypted_data_np = crypto.unsign(
        private_key_np,
        public_key_np,
        configuration_np,
    )
    decrypted_data = decrypted_data_np.tobytes()
    return json.loads(decrypted_data)


def encode_state_in_clear(state: State) -> bytes:
    json_data = json.dumps(state.to_dict())
    return json_data.encode()



from collections.abc import Callable
from typing import List, Any, Dict

from drax_sdk.broker.amqp_broker import DraxAmqpBroker
from drax_sdk.clients.drax_core_client import DraxCoreClient
from drax_sdk.clients.automation_client import AutomationClient
from drax_sdk.clients.data_miner_client import DataMinerClient
from drax_sdk.clients.alarms_client import AlarmsClient
from drax_sdk.model.config import DraxConfigParams
from drax_sdk.model.event import Event
from drax_sdk.model.node import NodeType, Node, State, Configuration
from drax_sdk.model.project import Project
from drax_sdk.model.automations import Automation
from drax_sdk.model.dto import (
    PagedResult,
    HandshakeRequest,
    HandshakeResponse,
    InstalledNode,
    FindNodeByIdsRequest,
    StateRequest,
    ConfigurationRequest,
    StateResponse,
    FlatConfigurationResponse,
    InstallRequest,
    PrepareRequest,
    # Automation models
    FindAutomationsRequest,
    FindUserActionsRequest,
    FindUserConditionsRequest,
    FindLogRequest,
    UserAction,
    UserCondition,
    LogEntry,
    # Alarm models
    Alarm,
    AlarmType,
    AlarmLogEntry,
    AcknowledgeAlarmRequest,
    ResolveAlarmRequest,
    FindAlarmLogEntriesRequest,
    FindAlarmTypesRequest,
    FindAlarmsRequest,
    # Data miner models
    DataRequest,
    DataResponse,
)
from drax_sdk.utils.codec import encode_state
from drax_sdk.utils.keystore import KeyStore


class DraxCore:

    client: DraxCoreClient
    config: DraxConfigParams

    def __init__(self, client: DraxCoreClient, config: DraxConfigParams):
        self.client = client
        self.config = config

    def register_node_type(self, node_type: NodeType) -> NodeType:
        return self.client.register_node_type(node_type)

    def update_node_type(self, node_type: NodeType) -> None:
        return self.client.update_node_type(node_type)

    def get_node_type_by_id(self, node_type_id: str) -> NodeType:
        return self.client.get_node_type_by_id(node_type_id)

    def unregister_node_type(self, node_type_id: str) -> None:
        self.client.unregister_node_type(node_type_id)

    # todo: da controllare perche il client non sembra avere parametri
    def list_node_types(self, project_id: str | None = None,) -> PagedResult[NodeType]:
        return self.client.list_node_types(project_id)

    def handshake(self, request: HandshakeRequest) -> HandshakeResponse:
        return self.client.handshake(request)

    def get_my_project(self) -> Project:
        return self.client.get_my_project()

    def get_project_by_id(self, id: str) -> Project:
        return self.client.get_project_by_id(id)

    def register_project(self, project: Project) -> Project:
        return self.client.register_project(project)

    def update_project(self, project: Project) -> None:
        self.client.update_project(project)

    def unregister_project(self, id: str) -> None:
        self.client.unregister_project(id)

    def prepare_node(self, request: PrepareRequest) -> InstalledNode:
        return self.client.prepare_node(request)

    def install_node(self, request: InstallRequest) -> InstalledNode:
        return self.client.install_node(request)

    # def generate_key_pair(self, request: ECDHKeysPairRequest) -> ECDHKeysPairResponse:
    #     return self.client.generate_keys_pair(request)
    #
    # def revoke_key_pair(self, request: ECDHRevokeRequest) -> ECDHRevokeResponse:
    #     return self.client.revoke_key(request)

    def update_node(self, node: Node) -> None:
        self.client.update_node(node)

    def uninstall_node(self, node_id: str) -> None:
        self.client.uninstall_node(node_id)

    def get_node_by_id(self, node_id: str) -> Node:
        return self.client.get_node_by_id(node_id)

    def get_nodes_by_ids(
        self, find_node_by_ids_request: FindNodeByIdsRequest
    ) -> List[Node]:
        return self.client.get_nodes_by_ids(find_node_by_ids_request)

    def list_projects(self, page: int = None, size: int = None) -> PagedResult[Project]:
        return self.client.list_projects()

    def list_nodes(
        self,
        project_id: str | None = None,
        keyword: str | None = None,
        page: int = 1,
        size: int = 10,
    ) -> PagedResult[Node]:
        return self.client.list_nodes(project_id, keyword, page, size)

    def list_states(
        self,
        node_id: str,
        project_id: str | None = None,
        from_time: int = None,
        to_time: int = None,
        page: int = 1,
        size: int = 10,
    ) -> PagedResult[StateResponse]:

        return self.client.list_states(
            node_id=node_id,
            project_id=project_id,
            from_time=from_time,
            to_time=to_time,
            page=page,
            size=size,
        )

    def list_configurations(
        self,
        node_id: str,
        from_time: int = None,
        to_time: int = None,
        page: int = 1,
        size: int = 10,
    ) -> PagedResult[FlatConfigurationResponse]:

        return self.client.list_configurations(
            node_id=node_id,
            from_time=from_time,
            to_time=to_time,
            page=page,
            size=size,
        )

    def list_nodes_states(
        self,
        find_node_by_ids_request: FindNodeByIdsRequest,
        start: int,
        end: int,
        page: int,
        size: int,
        project_id: str | None = None,
    ) -> PagedResult[StateResponse]:

        return self.client.list_nodes_states(
            project_id, find_node_by_ids_request, start, end, page, size
        )

    def set_state(self, node_id: str, state_request: StateRequest) -> None:
        self.client.set_state(node_id, state_request)

    def get_state(self, node_id: str) -> StateResponse:
        return self.client.get_state(node_id)

    def set_configuration(
        self, node_id: str, configuration_request: ConfigurationRequest
    ) -> None:
        # convert all values in str
        for key, value in configuration_request.configuration.items():
            configuration_request.configuration[key] = str(value)

        self.client.set_configuration(node_id, configuration_request)

    def invoke(self, event: Event) -> None:
        self.client.invoke(event)


class DraxAutomation:
    """Wrapper for AutomationClient following Java DraxAutomation pattern"""

    def __init__(self, client: AutomationClient):
        self.client = client

    def automate(
        self, activator_type: str, selector: str, data: dict = None
    ) -> None:
        self.client.automate(activator_type, selector, data)

    def delete(self, automation_id: str) -> None:
        self.client.delete(automation_id)

    def delete_user_action(self, user_action_id: str) -> None:
        self.client.delete_user_action(user_action_id)

    def delete_user_condition(self, user_condition_id: str) -> None:
        self.client.delete_user_condition(user_condition_id)

    def find(self, find_automations_request: FindAutomationsRequest) -> PagedResult[Automation]:
        return self.client.find(find_automations_request)

    def find_user_actions(
        self, find_user_actions_request: FindUserActionsRequest
    ) -> PagedResult[UserAction]:
        return self.client.find_user_actions(find_user_actions_request)

    def find_user_conditions(
        self, find_user_conditions_request: FindUserConditionsRequest
    ) -> PagedResult[UserCondition]:
        return self.client.find_user_conditions(find_user_conditions_request)

    def find_automation_log_entries(
        self, find_log_request: FindLogRequest
    ) -> PagedResult[LogEntry]:
        return self.client.find_logs(find_log_request)

    def get_actions(self, project_id: str) -> List[UserAction]:
        return self.client.get_actions(project_id)

    def get_by_id(self, automation_id: str) -> Automation:
        return self.client.get_by_id(automation_id)

    def get_conditions(self, project_id: str) -> List[UserCondition]:
        return self.client.get_conditions(project_id)

    def get_user_action_by_id(self, user_action_id: str) -> UserAction:
        return self.client.get_user_action_by_id(user_action_id)

    def get_user_condition_by_id(self, user_condition_id: str) -> UserCondition:
        return self.client.get_user_condition_by_id(user_condition_id)

    def save(self, automation: Automation) -> Automation:
        return self.client.save(automation)

    def save_user_action(self, user_action: UserAction) -> UserAction:
        return self.client.save_user_action(user_action)

    def save_user_condition(self, user_condition: UserCondition) -> UserCondition:
        return self.client.save_user_condition(user_condition)


class DraxDataMiner:
    """Wrapper for DataMinerClient following Java DraxDataMiner pattern"""

    def __init__(self, client: DataMinerClient):
        self.client = client

    def get_supported_aggregation_functions(self) -> List[str]:
        return self.client.get_supported_aggregation_functions()

    def get_supported_aggregation_periods(self) -> List[str]:
        return self.client.get_supported_aggregation_periods()

    def get_supported_time_windows(self) -> List[str]:
        return self.client.get_supported_time_windows()

    def fetch(self, request: DataRequest) -> DataResponse:
        return self.client.fetch(request)


class DraxAlarms:
    """Wrapper for AlarmsClient following Java DraxAlarms pattern"""

    def __init__(self, client: AlarmsClient):
        self.client = client

    def get_alarm_collection_strategy_types(self) -> List[str]:
        return self.client.get_alarm_collection_strategy_types()

    def get_alarm_scope_types(self) -> List[str]:
        return self.client.get_alarm_scope_types()

    def get_alarm_severity_types(self) -> List[str]:
        return self.client.get_alarm_severity_types()

    def get_alarm_status_types(self) -> List[str]:
        return self.client.get_alarm_status_types()

    def get_alarm_type_statuses(self) -> List[str]:
        return self.client.get_alarm_type_statuses()

    def acknowledge(self, acknowledge_alarm_request: AcknowledgeAlarmRequest) -> Alarm:
        return self.client.acknowledge(acknowledge_alarm_request)

    def delete(self, project_id: str, alarm_id: str) -> None:
        self.client.delete(project_id, alarm_id)

    def delete_alarm_type(self, alarm_type_id: str) -> None:
        self.client.delete_alarm_type(alarm_type_id)

    def find_alarm_log_entries(
        self, find_request: FindAlarmLogEntriesRequest
    ) -> PagedResult[AlarmLogEntry]:
        return self.client.find_alarm_log_entries(find_request)

    def find_alarm_types(self, find_alarm_types_request: FindAlarmTypesRequest) -> PagedResult[AlarmType]:
        return self.client.find_alarm_types(find_alarm_types_request)

    def find_alarms(self, find_alarms_request: FindAlarmsRequest) -> PagedResult[Alarm]:
        return self.client.find_alarms(find_alarms_request)

    def get_alarm(self, alarm_id: str) -> Alarm:
        return self.client.get_alarm(alarm_id)

    def get_alarm_type(self, alarm_type_id: str) -> AlarmType:
        return self.client.get_alarm_type(alarm_type_id)

    def resolve(self, resolve_alarm_request: ResolveAlarmRequest) -> Alarm:
        return self.client.resolve(resolve_alarm_request)

    def save_alarm_type(self, alarm_type: AlarmType) -> AlarmType:
        return self.client.save_alarm_type(alarm_type)


class Drax:
    def __init__(self, config: DraxConfigParams):
        self.config = config
        self.core = DraxCore(
            DraxCoreClient(config.drax_core_url, config.api_key, config.api_secret),
            config,
        )
        self.drax_automation = DraxAutomation(
            AutomationClient(config.automation_url, config.api_key, config.api_secret)
        )
        self.drax_data_miner = DraxDataMiner(
            DataMinerClient(config.data_miner_url, config.api_key, config.api_secret)
        )
        self.drax_alarms = DraxAlarms(
            AlarmsClient(config.alarms_url, config.api_key, config.api_secret)
        )
        self.broker = DraxAmqpBroker(config)

    def start(self):
        self.broker.start()
        pass

    def stop(self):
        self.broker.stop()
        pass

    def set_state(
        self,
        state: State | Dict[str, Any],
        node_id: str = None,
        cryptography_disabled=False,
        urn: str = None,
    ):
        if not isinstance(state, State):
            if node_id is None:
                raise ValueError("node_id must be provided")

        node_id = node_id if node_id else state.node_id
        if not node_id and not urn:
            raise ValueError("Either node_id or urn must be provided")

        node_private_key = KeyStore.get_private_key(node_id)
        
        # Get timestamp from state if it's a State object, otherwise use None
        timestamp = state.timestamp if isinstance(state, State) and hasattr(state, 'timestamp') else None

        request = StateRequest(
            node_id=node_id,
            state=encode_state(node_private_key, state),
            cryptography_disabled=cryptography_disabled,
            timestamp=timestamp,
            urn=urn,            
        )
        self.core.set_state(node_id, request)

    def set_configuration(
        self,
        configuration: Configuration | Dict[str, Any],
        node_id: str = None,
        cryptography_disabled=False,
        urn: str = None,
    ):
        if not isinstance(configuration, Configuration):
            if node_id is None:
                raise ValueError("node_id must be provided")

        node_id = node_id if node_id else configuration.node_id
        if not node_id and not urn:
            raise ValueError("Either node_id or urn must be provided")

        configuration_map = (
            configuration.to_map()
            if isinstance(configuration, Configuration)
            else configuration
        )

        self.core.set_configuration(
            node_id,
            ConfigurationRequest(
                node_id=node_id,
                configuration=configuration_map,
                cryptography_disabled=cryptography_disabled,
                urn=urn,
            ),
        )

    def add_configuration_listener(
        self, topic: str, listener: Callable[[Configuration], None]
    ):
        self.broker.add_configuration_listener(topic, listener)

    def add_state_listener(self, topic: str, listener: Callable[[State], None]):
        self.broker.add_state_listener(topic, listener)

    def add_events_listener(
        self, listener: Callable[[Event], None], project_id: str = None
    ):
        project_id = project_id or self.config.project_id
        if not project_id:
            raise ValueError("project_id is required")

        self.broker.add_event_listener(project_id=project_id, cb=listener)



import os
from typing import Optional

from pydantic import BaseModel


class DraxConfigParams(BaseModel):
    project_id: Optional[str] = None
    api_key: str
    api_secret: str
    drax_core_url: str
    automation_url: str
    broker_port: int
    broker_host: str
    broker_vhost: str
    data_miner_url: str
    alarms_url: str

    @classmethod
    def load_from_env(cls):
        return cls(
            project_id=os.getenv("DRAX_PROJECT_ID"),
            api_key=os.getenv("DRAX_PROJECT_API_KEY", "guest"),
            api_secret=os.getenv("DRAX_PROJECT__API_SECRET", "guest"),
            drax_core_url=os.getenv("DRAX_CORE_URL", "https://drax.network/core"),
            automation_url=os.getenv(
                "DRAX_AUTOMATION_URL", "https://drax.network/automation"
            ),
            broker_port=int(os.getenv("DRAX_BROKER_PORT", 5672)),
            broker_host=os.getenv("DRAX_BROKER_HOST", "35.205.187.28"),
            broker_vhost=os.getenv("DRAX_BROKER_VHOST", "/"),
            data_miner_url=os.getenv(
                "DRAX_DATA_MINER_URL", "https://drax.network/data-miner"
            ),
            alarms_url=os.getenv("DRAX_ALARMS_URL", "https://drax.network/alarms"),
        )

    @classmethod
    def standard(cls, project_id: str, api_key: str, api_secret: str):
        return cls(
            project_id=project_id,
            api_key=api_key,
            api_secret=api_secret,
            drax_core_url="https://drax.network/core",
            automation_url="https://drax.network/automation",
            data_miner_url="https://drax.network/data-miner",
            alarms_url="https://drax.network/alarms",
            broker_host="broker.drax.network",
            broker_port=5672,
            broker_vhost="/",
        )



from typing import List

import requests

from drax_sdk.model.automations import Automation
from drax_sdk.model.dto import (
    FindAutomationsRequest,
    FindUserActionsRequest,
    FindUserConditionsRequest,
    FindLogRequest,
    AutomationRequest,
    ExecuteAutomationRequest,
    UserAction,
    UserCondition,
    LogEntry,
    PagedResult,
)


class AutomationClient:
    DEFAULT_PAGE = 1
    DEFAULT_SIZE = 10

    service_url = ""
    api_key = ""
    api_secret = ""

    def __init__(self, service_url: str, api_key: str, api_secret: str):
        self.service_url = service_url
        self.api_key = api_key
        self.api_secret = api_secret

    def _get_headers(self):
        return {"drax-api-key": self.api_key, "drax-api-secret": self.api_secret}

    def automate(
        self, activator_type: str, selector: str, initial_variables: dict = None
    ) -> None:
        """Automate an activator"""
        request = AutomationRequest(
            activator_type=activator_type,
            selector=selector,
            initial_variables=initial_variables or {}
        )
        response = requests.post(
            self.service_url + "/automate",
            json=request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()

    def delete(self, automation_id: str) -> None:
        """Delete an automation"""
        response = requests.delete(
            self.service_url + f"/automations/{automation_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def delete_user_action(self, user_action_id: str) -> None:
        """Delete a user action"""
        response = requests.delete(
            self.service_url + f"/user-actions/{user_action_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def delete_user_condition(self, user_condition_id: str) -> None:
        """Delete a user condition"""
        response = requests.delete(
            self.service_url + f"/user-conditions/{user_condition_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()

    def find(self, find_request: FindAutomationsRequest) -> PagedResult[Automation]:
        """Find automations based on request"""
        response = requests.post(
            self.service_url + "/automations/find",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return PagedResult[Automation].model_validate(response.json())

    def find_user_actions(self, find_request: FindUserActionsRequest) -> PagedResult[UserAction]:
        """Find user actions based on request"""
        response = requests.post(
            self.service_url + "/user-actions/find",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return PagedResult[UserAction].model_validate(response.json())

    def find_user_conditions(self, find_request: FindUserConditionsRequest) -> PagedResult[UserCondition]:
        """Find user conditions based on request"""
        response = requests.post(
            self.service_url + "/user-conditions/find",
            json=find_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return PagedResult[UserCondition].model_validate(response.json())

    def get_actions(self, project_id: str) -> List[UserAction]:
        """Get actions for a project"""
        response = requests.get(
            self.service_url + f"/actions?projectId={project_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return [UserAction.model_validate(action) for action in response.json()]

    def get_by_id(self, automation_id: str) -> Automation:
        """Get an automation by id"""
        response = requests.get(
            self.service_url + f"/automations/{automation_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return Automation.model_validate(response.json())

    def get_conditions(self, project_id: str) -> List[UserCondition]:
        """Get conditions for a project"""
        response = requests.get(
            self.service_url + f"/conditions?projectId={project_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return [UserCondition.model_validate(cond) for cond in response.json()]

    def get_user_action_by_id(self, user_action_id: str) -> UserAction:
        """Get a user action by id"""
        response = requests.get(
            self.service_url + f"/user-actions/{user_action_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return UserAction.model_validate(response.json())

    def get_user_condition_by_id(self, user_condition_id: str) -> UserCondition:
        """Get a user condition by id"""
        response = requests.get(
            self.service_url + f"/user-conditions/{user_condition_id}",
            headers={**self._get_headers(), "Accept": "application/json"},
        )
        response.raise_for_status()
        return UserCondition.model_validate(response.json())

    def find_logs(self, log_request: FindLogRequest) -> PagedResult[LogEntry]:
        """Find automation log entries"""
        response = requests.post(
            self.service_url + "/logs/find",
            json=log_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return PagedResult[LogEntry].model_validate(response.json())

    def save(self, automation: Automation) -> Automation:
        """Save an automation"""
        response = requests.post(
            self.service_url + "/automations",
            json=automation.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return Automation.model_validate(response.json())

    def save_user_action(self, user_action: UserAction) -> UserAction:
        """Save a user action"""
        response = requests.post(
            self.service_url + "/user-actions",
            json=user_action.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return UserAction.model_validate(response.json())

    def save_user_condition(self, user_condition: UserCondition) -> UserCondition:
        """Save a user condition"""
        response = requests.post(
            self.service_url + "/user-conditions",
            json=user_condition.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return UserCondition.model_validate(response.json())

    def execute_automation(self, execute_request: ExecuteAutomationRequest) -> Automation:
        """Execute an automation"""
        response = requests.post(
            self.service_url + "/automations/execute",
            json=execute_request.model_dump(by_alias=True),
            headers={**self._get_headers(), "Content-Type": "application/json"},
        )
        response.raise_for_status()
        return Automation.model_validate(response.json())



from typing import Dict

from pydantic import BaseModel, Field

from drax_sdk.model.node import Node
from drax_sdk.utils.timestamp import unix_timestamp


class Event(BaseModel):
    timestamp: int = Field(default_factory=unix_timestamp)
    node: Node | None = None
    project_id: str | None = Field(alias="projectId", default=None)
    type: str
    values: Dict[str, str] | None = None

    @classmethod
    def create(cls, type: str, project_id: str, node: Node, params: Dict[str, str]):
        event = cls(type=type, project_id=project_id, node=node)
        for key, value in params.items():
            setattr(event, key, value)
        return event

    class Config:
        populate_by_name = True
        extra = "allow"



from typing import List

from pydantic import BaseModel

from drax_sdk.utils.timestamp import unix_timestamp


class Value(BaseModel):
    name: str
    value: str


class BaseValueObject(BaseModel):
    values: List[Value] = []

    def all_names(self):
        return [value.name for value in self.values]

    def all_values(self):
        return self.values

    def get_value(self, name: str):
        for value in self.values:
            if value.name == name:
                return value.value
        return None

    def set_value(self, name: str, value: str):
        for val in self.values:
            if val.name == name:
                val.value = value
                return
        self.values.append(Value(name=name, value=value))

    def remove_value(self, name: str):
        self.values = [value for value in self.values if value.name != name]

    def clear_values(self):
        self.values.clear()

    @classmethod
    def from_dict(cls, map: dict):
        return cls(
            timestamp=unix_timestamp(),
            values=[Value(name=k, value=str(v)) for k, v in map.items()],
        )

    def to_dict(self):
        return {value.name: value.value for value in self.values}

    @staticmethod
    def merge(*measures):
        merged = BaseValueObject(values=[])
        for measure in measures:
            if measure is not None:
                for value in measure.values:
                    merged.set_value(value.name, value.value)
        return merged


class NodeEntry(BaseValueObject):
    node_id: str | None = None
    timestamp: int | None = 0



from typing import List, Optional, Any

from pydantic import BaseModel, Field

from drax_sdk.model.dynamic import NodeEntry
from drax_sdk.model.utils import BytesBase64Model
from drax_sdk.utils.timestamp import unix_timestamp


class ECDHKeyType:
    ACTIVE = "active"
    DEFAULT = "default"
    REVOKED = "revoked"


class ECDHKey(BytesBase64Model):
    key: bytes
    status: str


class State(NodeEntry):

    def merge(self, node_entry: NodeEntry):
        for value in node_entry.all_values():
            self.set_value(value.name, value.value)


class Configuration(NodeEntry):

    def merge(self, node_entry: NodeEntry):
        for value in node_entry.all_values():
            self.set_value(value.name, value.value)


class Property(BaseModel):
    name: str
    type: str


class Descriptor(BaseModel):
    properties: List[Property]

    def add_property(self, name: str, type: str):
        if not any(p.name == name for p in self.properties):
            self.properties.append(Property(name=name, type=type))


class NodeType(BaseModel):
    id: str
    name: str
    tag: Optional[str] = Field(default=None)
    project_id: str | None = Field(alias="projectId", default=None)
    peripherals: Descriptor

    def supports(self, property_name: str):
        return any(p == property_name for p in self.peripherals.properties)

    class Config:
        populate_by_name = True


class Value(BaseModel):
    name: str
    value: str


class Node(BaseModel):
    id: Optional[str] = Field(None)
    association_code: Optional[str] = Field(alias="associationCode", default=None)
    project_id: Optional[str] = Field(None, alias="projectId")
    name: str
    urn: Optional[str] = None
    tag: Optional[str] = None
    supported_types: List[str] = Field(default_factory=set, alias="supportedTypes")
    state_publish_topic: Optional[str] = Field(None, alias="statePublishTopic")
    configuration_publish_topic: Optional[str] = Field(
        None, alias="configurationPublishTopic"
    )
    public_keys: List[ECDHKey] = Field(default_factory=list, alias="publicKeys")
    active: bool = False
    script: Optional[str] = None
    max_idle_time: int = Field(0, alias="maxIdleTime")
    last_check: int = Field(default=unix_timestamp(), alias="lastCheck")
    status: Optional[str] = None
    state: State = State()
    values: List[Value] = Field(default_factory=list)

    def add_supported_type(self, type: str):
        self.supported_types.append(type)

    def add_public_key(self, public_key: bytes):
        if not any(k.key == public_key for k in self.public_keys):
            self.public_keys.append(ECDHKey(key=public_key, status=ECDHKeyType.ACTIVE))

    def supports(self, type: str) -> bool:
        return type in self.supported_types

    def revoke_public_key(self, public_key: bytes):
        for key in self.public_keys:
            if key.key == public_key:
                key.status = ECDHKeyType.REVOKED

    def is_key_enabled(self, public_key: bytes) -> bool:
        return any(
            k.key == public_key
            and k.status in (ECDHKeyType.ACTIVE, ECDHKeyType.DEFAULT)
            for k in self.public_keys
        )

    def set_single_public_key(self, public_key: bytes):
        self.public_keys = [ECDHKey(key=public_key, status=ECDHKeyType.DEFAULT)]

    def get_default_key(self) -> Optional[bytes]:
        for key in self.public_keys:
            if key.status == ECDHKeyType.DEFAULT:
                return key.key
        for key in self.public_keys:
            if key.status == ECDHKeyType.ACTIVE:
                return key.key
        return None

    def set_value(self, name: str, value: Any):
        self.values.append(Value(name=name, value=str(value)))

    def get_value(self, name: str) -> str | None:
        for value in self.values:
            if value.name == name:
                return value.value
        return None

    class Config:
        populate_by_name = True



from typing import List

import requests

from drax_sdk.model.dto import DataRequest, DataResponse


class DataMinerClient:
    service_url = ""
    api_key = ""
    api_secret = ""

    def __init__(self, service_url: str, api_key: str, api_secret: str):
        self.service_url = service_url
        self.api_key = api_key
        self.api_secret = api_secret

    def _get_headers(self):
        return {"drax-api-key": self.api_key, "drax-api-secret": self.api_secret}

    def get_supported_aggregation_functions(self) -> List[str]:
        """Get supported aggregation functions"""
        response = requests.get(
            self.service_url + "/aggregation-functions",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_supported_aggregation_periods(self) -> List[str]:
        """Get supported aggregation periods"""
        response = requests.get(
            self.service_url + "/aggregation-periods",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def get_supported_time_windows(self) -> List[str]:
        """Get supported time windows"""
        response = requests.get(
            self.service_url + "/time-windows",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return response.json()

    def fetch(self, data_request: DataRequest) -> DataResponse:
        """Fetch data based on request"""
        response = requests.post(
            self.service_url + "/data",
            json=data_request.model_dump(by_alias=True),
            headers={
                **self._get_headers(),
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        response.raise_for_status()
        return DataResponse.model_validate(response.json())



from typing import Optional, List, TypeVar, Generic, Dict, Set, Any
from enum import Enum

from pydantic import Field, BaseModel

from drax_sdk.model.dynamic import Value
from drax_sdk.model.node import State
from drax_sdk.model.utils import BytesBase64Model
from drax_sdk.utils.timestamp import unix_timestamp

T = TypeVar("T")


class PagedResult(BaseModel, Generic[T]):
    results: List[T] = []
    total_rows: int = Field(alias="totalRows")

    class Config:
        populate_by_name = True


class HandshakeRequest(BaseModel):
    node_id: int | None = Field(alias="nodeId", default=None)
    name: str | None = None
    association_code: str | None = Field(alias="associationCode", default=None)
    urn: str | None = None
    project_id: str | None = Field(alias="projectId", default=None)
    supported_types: List[str] | None = Field(alias="supportedTypes", default=None)
    configuration_publish_topic: str | None = Field(
        alias="configurationPublishTopic", default=None
    )
    state_publish_topic: str | None = Field(alias="statePublishTopic", default=None)
    initial_state: State | None = Field(alias="initialState", default=None)
    extras: List[Value] | None = None

    class Config:
        populate_by_name = True


class HandshakeResponse(BaseModel):
    node_id: int = Field(alias="nodeId")
    urn: str
    public_key: bytes | None = Field(alias="publicKey")
    private_key: bytes | None = Field(alias="privateKey")

    class Config:
        populate_by_name = True


class AssociationRequest(BaseModel):
    apiKey: str
    apiSecret: str
    associationCode: str
    urn: str


class ConfigurationRequest(BaseModel):
    api_key: str | None = Field(alias="apiKey", default=None)
    api_secret: str | None = Field(alias="apiSecret", default=None)
    node_id: Optional[int] = Field(alias="nodeId", default=None)
    urn: str | None = None
    codec: str | None = None
    cryptography_disabled: bool = Field(alias="cryptographyDisabled", default=False)
    timestamp: int = Field(alias="timestamp", default_factory=unix_timestamp)
    configuration: Dict[str, str]

    @classmethod
    def from_configuration(cls, configuration):
        return cls(
            node_id=configuration.node_id,
            timestamp=configuration.timestamp,
            configuration=configuration.to_map(),
        )

    class Config:
        populate_by_name = True


class ConfigurationResponse(BytesBase64Model):
    node_id: str = Field(alias="nodeId")
    timestamp: int = Field(alias="timestamp")
    urn: str | None = Field(alias="urn", default=None)
    configuration: bytes = Field(alias="configuration")
    cryptography_disabled: bool = Field(alias="cryptographyDisabled", default=False)

    class Config:
        populate_by_name = True


class PrepareRequest(BaseModel):
    association_code: Optional[str] = Field(alias="associationCode", default=None)
    project_id: Optional[str] = Field(None, alias="projectId")
    name: str
    urn: Optional[str] = None
    tag: Optional[str] = None
    supported_types: List[str] = Field(default_factory=set, alias="supportedTypes")
    state_publish_topic: Optional[str] = Field(None, alias="statePublishTopic")
    configuration_publish_topic: Optional[str] = Field(
        None, alias="configurationPublishTopic"
    )
    script: Optional[str] = None
    max_idle_time: int = Field(0, alias="maxIdleTime")
    last_check: int = Field(default=unix_timestamp(), alias="lastCheck")
    state: State = State()

    class Config:
        populate_by_name = True


class InstallRequest(BaseModel):
    association_code: Optional[str] = Field(alias="associationCode", default=None)
    project_id: Optional[str] = Field(None, alias="projectId")
    name: str
    urn: Optional[str] = None
    tag: Optional[str] = None
    supported_types: List[str] = Field(default_factory=set, alias="supportedTypes")
    state_publish_topic: Optional[str] = Field(None, alias="statePublishTopic")
    configuration_publish_topic: Optional[str] = Field(
        None, alias="configurationPublishTopic"
    )
    script: Optional[str] = None
    max_idle_time: int = Field(0, alias="maxIdleTime")
    last_check: int = Field(default=unix_timestamp(), alias="lastCheck")
    state: State = State()

    class Config:
        populate_by_name = True


class FlatConfigurationResponse(BaseModel):
    node_id: Optional[str] = Field(alias="nodeId")
    urn: Optional[str] = None
    timestamp: int
    configuration: Dict[str, str]

    @classmethod
    def from_node_and_entry(cls, node, state):
        return cls(
            nodeId=node.id,
            urn=node.urn,
            timestamp=state.timestamp,
            state=state.to_map(),
        )

    @classmethod
    def from_state(cls, state):
        return cls(nodeId=state.nodeId, timestamp=state.timestamp, state=state.to_map())


class StateRequest(BytesBase64Model):
    api_key: str = Field(alias="apiKey", default=None)
    api_secret: str = Field(alias="apiSecret", default=None)
    node_id: Optional[str] = Field(alias="nodeId", default=None)
    urn: Optional[str] = None
    timestamp: Optional[int] = None
    state: bytes
    codec: Optional[str] = None
    cryptography_disabled: bool = Field(alias="cryptographyDisabled", default=False)

    class Config:
        populate_by_name = True


class StateResponse(BaseModel):
    node_id: Optional[str] = Field(alias="nodeId")
    urn: Optional[str] = None
    timestamp: int
    state: Dict[str, str]

    @classmethod
    def from_node_and_entry(cls, node, state):
        return cls(
            nodeId=node.id,
            urn=node.urn,
            timestamp=state.timestamp,
            state=state.to_map(),
        )

    @classmethod
    def from_state(cls, state):
        return cls(nodeId=state.nodeId, timestamp=state.timestamp, state=state.to_map())


class AuthenticationRequest(BaseModel):
    api_key: str = Field(alias="apiKey")
    api_secret: str = Field(alias="apiSecret")

    class Config:
        populate_by_name = True


class FindNodeByIdsRequest(BaseModel):
    node_ids: List[int] = Field(alias="nodeIds")

    class Config:
        populate_by_name = True


from pydantic import Field
from typing import Optional


class InstalledNode(BytesBase64Model):
    id: str
    urn: str
    public_key: Optional[bytes] = Field(None, alias="publicKey")
    private_key: Optional[bytes] = Field(None, alias="privateKey")

    @property
    def public_key_hex(self) -> str:
        if self.public_key is None:
            return ""
        return self.public_key.hex()

    @property
    def private_key_hex(self) -> str:
        if self.private_key is None:
            return ""
        return self.private_key.hex()

    class Config:
        populate_by_name = True


# =============================================================================
# ALARM MODELS AND REQUESTS
# =============================================================================

class AlarmSeverity(str, Enum):
    """Alarm severity levels."""
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class AlarmStatus(str, Enum):
    """Alarm status types."""
    INACTIVE = "INACTIVE"
    PENDING = "PENDING"
    TRIGGERED = "TRIGGERED"
    ACKNOWLEDGED = "ACKNOWLEDGED"
    RESOLVED = "RESOLVED"


class AlarmTypeStatus(str, Enum):
    """Alarm type status."""
    ARMED = "ARMED"
    DISARMED = "DISARMED"
    AUTO = "AUTO"


class ScopeType(str, Enum):
    """Scope type for alarms."""
    LOCAL = "LOCAL"
    GLOBAL = "GLOBAL"
    TAGS = "TAGS"
    NODE_LIST = "NODE_LIST"


class Scope(BaseModel):
    """Alarm scope definition."""
    type: ScopeType = Field(default=ScopeType.GLOBAL)
    node_ids: Optional[List[str]] = Field(default=None, alias="nodeIds")
    tags: Optional[List[str]] = None

    class Config:
        populate_by_name = True


class AlarmType(BaseModel):
    """Alarm type definition."""
    id: Optional[str] = None
    name: str
    project_id: str = Field(alias="projectId")
    description: Optional[str] = None
    severity: AlarmSeverity
    scope: Scope
    status: AlarmTypeStatus
    
    class Config:
        populate_by_name = True


class Alarm(BaseModel):
    """Alarm instance."""
    id: Optional[str] = None
    alarm_type_id: str = Field(alias="alarmTypeId")
    project_id: str = Field(alias="projectId")
    status: AlarmStatus
    severity: AlarmSeverity
    triggered_at: Optional[int] = Field(default=None, alias="triggeredAt")
    acknowledged_at: Optional[int] = Field(default=None, alias="acknowledgedAt")
    resolved_at: Optional[int] = Field(default=None, alias="resolvedAt")
    duration: Optional[int] = None
    node_id: Optional[str] = Field(default=None, alias="nodeId")
    message: Optional[str] = None

    class Config:
        populate_by_name = True


class AlarmLogEntry(BaseModel):
    """Alarm log entry."""
    id: Optional[str] = None
    alarm_id: str = Field(alias="alarmId")
    project_id: str = Field(alias="projectId")
    timestamp: int
    message: str
    level: str

    class Config:
        populate_by_name = True


class AcknowledgeAlarmRequest(BaseModel):
    """Request to acknowledge an alarm."""
    alarm_id: str = Field(alias="alarmId")
    project_id: str = Field(alias="projectId")

    class Config:
        populate_by_name = True


class ResolveAlarmRequest(BaseModel):
    """Request to resolve an alarm."""
    alarm_id: str = Field(alias="alarmId")
    project_id: str = Field(alias="projectId")

    class Config:
        populate_by_name = True


class GetAlarmRequest(BaseModel):
    """Request to get an alarm by ID."""
    id: str

    class Config:
        populate_by_name = True


class GetAlarmTypeRequest(BaseModel):
    """Request to get an alarm type by ID."""
    id: str

    class Config:
        populate_by_name = True


class PageableRequest(BaseModel):
    """Base class for pageable requests."""
    page: int = Field(default=1)
    size: int = Field(default=10)

    class Config:
        populate_by_name = True


class FindAlarmLogEntriesRequest(PageableRequest):
    """Request to find alarm log entries."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    alarm_id: Optional[str] = Field(default=None, alias="alarmId")

    class Config:
        populate_by_name = True


class FindAlarmTypesRequest(PageableRequest):
    """Request to find alarm types."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    name: Optional[str] = None

    class Config:
        populate_by_name = True


class FindAlarmsRequest(PageableRequest):
    """Request to find alarms."""
    keyword: Optional[str] = None
    project_id: Optional[str] = Field(default=None, alias="projectId")
    statuses: Optional[List[AlarmStatus]] = None
    alarm_type_ids: Optional[List[str]] = Field(default=None, alias="alarmTypeIds")
    start: Optional[int] = None
    end: Optional[int] = None
    min_duration: Optional[int] = Field(default=None, alias="minDuration")
    max_duration: Optional[int] = Field(default=None, alias="maxDuration")
    node_ids: Optional[List[str]] = Field(default=None, alias="nodeIds")
    sorts: Optional[List[Dict[str, Any]]] = None

    class Config:
        populate_by_name = True


# =============================================================================
# AUTOMATION MODELS AND REQUESTS
# =============================================================================

class LogLevel(str, Enum):
    """Log level for automation logs."""
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"


class AutomationPageableRequest(BaseModel):
    """Base class for pageable requests (automation uses page=0 default)."""
    page: int = Field(default=0)
    size: int = Field(default=10)

    class Config:
        populate_by_name = True


class FindAutomationsRequest(AutomationPageableRequest):
    """Request to find automations."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    keyword: Optional[str] = None
    activator_type: Optional[str] = Field(default=None, alias="activatorType")

    class Config:
        populate_by_name = True


class FindUserActionsRequest(AutomationPageableRequest):
    """Request to find user actions."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    keyword: Optional[str] = None

    class Config:
        populate_by_name = True


class FindUserConditionsRequest(AutomationPageableRequest):
    """Request to find user conditions."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    keyword: Optional[str] = None

    class Config:
        populate_by_name = True


class FindLogRequest(BaseModel):
    """Request to find automation logs."""
    project_id: Optional[str] = Field(default=None, alias="projectId")
    automation_id: Optional[str] = Field(default=None, alias="automationId")
    level: Optional[LogLevel] = None
    keyword: Optional[str] = None
    from_time: int = Field(default=0, alias="from")
    to_time: int = Field(default=0, alias="to")
    sorts: Optional[List[Dict[str, Any]]] = None
    paging_state: Optional[Dict[str, Any]] = Field(default=None, alias="pagingState")

    class Config:
        populate_by_name = True


class AutomationRequest(BaseModel):
    """Request to trigger automation."""
    activator_type: str = Field(alias="activatorType")
    selector: str
    initial_variables: Dict[str, Any] = Field(default_factory=dict, alias="initialVariables")
    project_id: Optional[str] = Field(default=None, alias="projectId")

    class Config:
        populate_by_name = True


class ExecuteAutomationRequest(BaseModel):
    """Request to execute an automation."""
    automation_id: str = Field(alias="automationId")

    class Config:
        populate_by_name = True


class UserVariable(BaseModel):
    """User variable definition."""
    name: str
    label: str
    input_type: str = Field(alias="inputType")
    default_value: Optional[str] = Field(default=None, alias="defaultValue")

    class Config:
        populate_by_name = True


class UserAction(BaseModel):
    """User-defined action."""
    id: Optional[str] = None
    project_id: str = Field(alias="projectId")
    type: str
    description: str
    priority: int = 0
    variables: List[UserVariable] = Field(default_factory=list)

    class Config:
        populate_by_name = True


class UserCondition(BaseModel):
    """User-defined condition."""
    id: Optional[str] = None
    project_id: str = Field(alias="projectId")
    type: str
    description: str
    priority: int = 0
    variables: List[UserVariable] = Field(default_factory=list)
    script: Optional[str] = None

    class Config:
        populate_by_name = True


class LogEntry(BaseModel):
    """Automation log entry."""
    id: Optional[str] = None
    project_id: str = Field(alias="projectId")
    automation_id: str = Field(alias="automationId")
    message: str
    level: LogLevel
    timestamp: int

    class Config:
        populate_by_name = True


# =============================================================================
# DATA MINER MODELS AND REQUESTS
# =============================================================================

class AggregationFunction:
    """Supported aggregation functions."""
    NONE = "none"
    SUM = "sum"
    MEAN = "mean"
    MIN = "min"
    MAX = "max"
    STDDEV = "stddev"
    COUNT = "count"
    INTEGRAL = "integral"
    MEDIAN = "median"
    DISTINCT = "distinct"
    MODE = "mode"

    @staticmethod
    def values() -> List[str]:
        return [
            AggregationFunction.NONE,
            AggregationFunction.SUM,
            AggregationFunction.MEAN,
            AggregationFunction.MIN,
            AggregationFunction.MAX,
            AggregationFunction.STDDEV,
            AggregationFunction.COUNT,
            AggregationFunction.INTEGRAL,
            AggregationFunction.MEDIAN,
            AggregationFunction.DISTINCT,
            AggregationFunction.MODE,
        ]


class AggregationPeriod:
    """Supported aggregation periods."""
    LAST_MINUTE = "last-minute"
    LAST_HOUR = "last-hour"
    LAST_DAY = "last-day"
    LAST_WEEK = "last-week"
    LAST_MONTH = "last-month"
    LAST_THREE_MONTHS = "last-three-months"

    @staticmethod
    def values() -> List[str]:
        return [
            AggregationPeriod.LAST_MINUTE,
            AggregationPeriod.LAST_HOUR,
            AggregationPeriod.LAST_DAY,
            AggregationPeriod.LAST_WEEK,
            AggregationPeriod.LAST_MONTH,
            AggregationPeriod.LAST_THREE_MONTHS,
        ]


class TimeWindow:
    """Supported time windows."""
    SECONDS_1 = "1s"
    SECONDS_5 = "5s"
    SECONDS_10 = "10s"
    SECONDS_30 = "30s"
    MINUTES_1 = "1m"
    MINUTES_30 = "30m"
    HOURS_1 = "1h"
    HOURS_8 = "8h"
    DAYS_1 = "1d"

    @staticmethod
    def values() -> List[str]:
        return [
            TimeWindow.SECONDS_1,
            TimeWindow.SECONDS_5,
            TimeWindow.SECONDS_10,
            TimeWindow.SECONDS_30,
            TimeWindow.MINUTES_1,
            TimeWindow.MINUTES_30,
            TimeWindow.HOURS_1,
            TimeWindow.HOURS_8,
            TimeWindow.DAYS_1,
        ]


class DataRequest(BaseModel):
    """Request for data mining."""
    project_id: str = Field(alias="projectId")
    node_ids: Optional[List[str]] = Field(default=None, alias="nodeIds")
    peripherals: Optional[List[str]] = None
    tag: Optional[str] = None
    window: Optional[int] = None
    unit: str = "millisecond"
    from_time: Optional[int] = Field(default=None, alias="from")
    to_time: Optional[int] = Field(default=None, alias="to")
    aggregation_function: Optional[str] = Field(default=None, alias="aggregationFunction")
    group_nodes: bool = Field(default=False, alias="groupNodes")

    class Config:
        populate_by_name = True


class DataResponse(BaseModel):
    """Response from data mining."""
    rows: List[List[Any]]
    response_code: int = Field(alias="responseCode")
    peripherals: List[str]
    nodes: Set[str]

    class Config:
        populate_by_name = True


# =============================================================================
# CORE/HANDSHAKE MODELS AND REQUESTS (Original dto.py models)
# =============================================================================
